/* miditoweb - a MIDI-WebAPI translator to be used with Avolites Titan

miditoweb is released under the New BSD license (see LICENSE.txt). Go to the project
home page for more info: http://www.avolites.de/downloads/miditoweb

miditoweb uses inih to read/parse the config file.
inih is released under the New BSD license (see LICENSE.txt). Go to the project
home page for more info: https://github.com/benhoyt/inih

*/
#include <string>
#include <Windows.h>
#include <WinInet.h>
#include <iostream>
#include <conio.h>     /* include for kbhit() and getch() functions */
#include <stdio.h>
#include <mmsystem.h>
#include <cstdlib>
#include <sstream>
#include "INIReader.h"
#include <vector>
#include <unistd.h> // for access()

#define  INTERNET_OPEN_TYPE_DIRECT 1
#define  INTERNET_SERVICE_HTTP 3
#define GET 0
#define POST 1

#pragma comment (lib, "wininet.lib")

using namespace std;
string HttpRequest(string site, string param);
boolean debug = false;
boolean sendRequest = true;
string path;

bool DoesFileExist (const std::string& name) {
    return ( access( name.c_str(), F_OK ) != -1 );
}

std::string sections(INIReader &reader)
{
    std::stringstream ss;
    std::set<std::string> sections = reader.Sections();
    for (std::set<std::string>::iterator it = sections.begin(); it != sections.end(); ++it)
        ss << *it << ",";
    return ss.str();
}

string HttpRequest(string site, string param) {
	HINTERNET hInternet = InternetOpenW(L"MIDI Translator", INTERNET_OPEN_TYPE_DIRECT, NULL, NULL, 0);
	if (debug) cout << endl << "Debug HttpRequest()" << endl; 
	wstring widestr;
	for (int i = 0; i < site.length(); ++i)
		{
			widestr += wchar_t(site[i]);
		}
	const wchar_t* site_name = widestr.c_str();
		
	wstring widestr2;
	for (int i = 0; i < param.length(); ++i)
		{
			widestr2 += wchar_t(param[i]);
		}
	const wchar_t* site_param = widestr2.c_str();

	// convert str to const wchar_t as the args require!

	HINTERNET hConnect = InternetConnectW(hInternet,site_name, 4430, NULL, NULL, INTERNET_SERVICE_HTTP, 0, INTERNET_NO_CALLBACK);

	const wchar_t* parrAcceptTypes[] = { L"text/*", NULL }; 				// accepted types. 

	HINTERNET hRequest = HttpOpenRequestW(hConnect, L"GET", site_param, NULL, NULL, parrAcceptTypes, 0, 0);
	BOOL bRequestSent = HttpSendRequestW(hRequest, NULL, 0, NULL, 0);
	std::string strResponse;
	InternetCloseHandle(hRequest);
	InternetCloseHandle(hConnect);
	InternetCloseHandle(hInternet);
	return strResponse;
}

void CALLBACK midiCallback(HMIDIIN handle, UINT uMsg, DWORD dwInstance, DWORD dwParam1, DWORD dwParam2)
{
	INIReader reader(path);
    switch ( uMsg )
    {
    case MIM_OPEN:
         // cout << "-----OPENED.-----" << endl;
         break;
    case MIM_CLOSE:
         cout << "-----EVERYTHING IS CLOSING.-----" << endl;
         break;
    case MIM_DATA:
         // cout << "-----APPARENTLY THERE IS DATA.-----" << endl;
         break;
    case MIM_LONGDATA:
         cout << "-----LONGDATA'D.-----" << endl;
         break;
    case MIM_ERROR:
         cout << "-----ERROR.-----" << endl;
         break;
    case MIM_LONGERROR:
         cout << "-----LONGERROR.  EVEN WORSE.-----" << endl;
         break;
    }
	
	unsigned int byteArray[4];  										// this will hold the bytes received
	float val = 0;														// the value 0...1
	int note;															// the MIDI note
	int channel;														// the MIDI channel
	 
	byteArray[0] = (int)((dwParam1 >> 24) & 0xFF) ;
	byteArray[1] = (int)((dwParam1 >> 16) & 0xFF) ;						// the note
	byteArray[2] = (int)((dwParam1 >> 8) & 0XFF);						// the value
	byteArray[3] = (int)((dwParam1 & 0XFF));
	
	if (debug) {
		cout << endl << "Debug MIDICallback:" << endl;
		cout << "  dwInstance: " << dwInstance << endl;  
		cout << "  Handle: " << handle << endl;
		cout << "  dwParam1 (Raw Data): " << dwParam1 << endl; 			// dwParam1 is the bytes of the MIDI Message packed into an unsigned long
		cout << "  byteArray[0]: " << byteArray[0] << endl;				// always 0 as MIDI has only 3 bytes
		cout << "  byteArray[1]: " << byteArray[1] << endl;				// CC: the value, 0..127  PC: empty, Note: 127 [on] or 0 [off]
		cout << "  byteArray[2]: " << byteArray[2] << endl;				// CC: the note, PC: the value, Note: the note
		cout << "  byteArray[3]: " << byteArray[3] << endl;				// the status byte. 
																		// CC ch 0 is 176, CC ch 1 is 177, CC ch 15 is 191
																		// PC ch 0 is 192, PC ch 1 is 193, PC ch 15 is 207
																		// PC not implemented as it involves multiple byte sequences
																		// Note ch 0 is 144, ch 1 is 145, ch 15 is 159
																		
		cout << "  dwParam2 (timestamp): " << dwParam2 << endl; 		// dwParam2 is the timestamp of key press
		cout << "  uMsg:" << uMsg << endl;
		cout << "  ---------------------------------------" << endl;
	}
	
	if (((127 < byteArray[3]) && (byteArray[3] < 160)) || ((175 < byteArray[3]) && (byteArray[3] < 192))){
	
		note = byteArray[2];
		val = byteArray[1]/127.000;
		
		std::string notestring;  										// the string to lookup in the notes section
		
		if ((127 < byteArray[3]) && (byteArray[3] < 144)){				// check statusbyte for Note off
			channel = byteArray[3] - 127;
			notestring = "NOf.";
			if (debug) cout << "Ch. " << channel << " Note Off " << notestring << endl;
		}
		
																		// Note Off can either be the designated status byte (this does the Korg Nanokontrol)
																		// or it can be a Note On with valovity = 0 (this does the Behringer BCF2000)
		
		if ((143 < byteArray[3]) && (byteArray[3] < 160)){				// check statusbyte for Note on/off
			channel = byteArray[3] - 143;
			notestring = "NOn.";
			if (byteArray[1] == 0) notestring = "NOf.";
			if (debug) cout << "Ch. " << channel << " Note On/Off " << notestring << endl;
		}
		
		if ((175 < byteArray[3]) && (byteArray[3] < 192)){				// check statusbyte for Control Change
			channel = byteArray[3] - 175;
			notestring = "CCh.";
			if (debug) cout << "Ch. " << channel << " Control Change " << endl;
		}
		
		if (reader.GetInteger("general", "omni_ch", 0) > 0){			// if omni_ch set, redirect all MIDI commands to this channel
			channel = reader.GetInteger("general", "omni_ch", 0);
			if (debug) cout << "omni_ch:  " << reader.GetInteger("general", "omni_ch", 0) << " all mapping redirected to this channel" << endl;
		} 
		
		notestring.append(to_string(channel));							// build search string
		notestring.append(".");
		notestring.append(to_string(note));
		if (debug) cout << endl << "Searching for " << notestring << endl;
		
		std::string command;											// the API command string
		std::string mapping;											// the mapping as in the config
		
		vector<int> mappingvect;
		vector<string> commandvect;
	
		
		mapping = reader.Get("mapping", notestring, "empty");			// get mapping for this note from config
		std::stringstream mappingstring(mapping);						// mapping as stringstream
		int i;															// iterator for tokenizer
		int stringindex;												// the string index

		while (mappingstring >> i)										// if there is a mapping, tokenize it into mappingvect
		{
			mappingvect.push_back(i);
			if (mappingstring.peek() == ',')
				mappingstring.ignore();
		}
		
		if (mappingvect.size() == 2){									// a correct mapping has two elements
			stringindex = mappingvect[0];								// get stringindex from mapping
			std::string stringindexstring = std::to_string(stringindex);  	// string index as string
			command = reader.Get("strings", stringindexstring, "");		// get command from config

			if (command.size() > 20){									// proper commands are long strings
													
																		// now we have the string. next: parse for $arg and $lev
													
				if (command.find("$arg") != std::string::npos) command.replace(command.find("$arg"),4,to_string(mappingvect[1]));
				if (command.find("$lev") != std::string::npos) command.replace(command.find("$lev"),4,to_string(val));
				
				if (debug) cout << "HttpRequest: "<< endl << command << endl;
				
				if (sendRequest) HttpRequest(reader.Get("general", "ip", "127.0.0.1"), command);
			}
		}
	}
}

void MidiReceiver(){
	
	INIReader reader(path);
    MIDIINCAPS     mic;       
    unsigned long result;
    HMIDIIN      inHandle;
	string		inputstring;
	int			devid=0;
     
    unsigned long    iNumDevs, i;
    iNumDevs = midiInGetNumDevs();  									// number of MIDI In devices in this computer
 
 	if (iNumDevs < 1 ){
		std::cout << "No MIDI device detected, press q to close program."  << endl << endl;

    } else {
 
		for (i = 0; i < iNumDevs; i++)
		{
			if (!midiInGetDevCaps(i, &mic, sizeof(MIDIINCAPS)))
			{
				printf("Device ID #%u: %s\r\n", i, mic.szPname);			// show device ID and name
			}
		}
		
		cout << "Enter id of MIDI device to use [ENTER for ID #0]: ";
		getline (cin, inputstring);
		stringstream(inputstring) >> devid;
		cout << "Using MIDI device #" << devid << endl;
		
		result = midiInOpen(&inHandle, devid, (DWORD)midiCallback, 0, CALLBACK_FUNCTION);
		
		if (result)
		{
		   cout << "There was an error opening the MIDI In device. " << endl;
		   cout << "Is another program listening to MIDI?" << endl;
		   cout << "Press q to close program." << endl << endl;
		}
		else
		{
			midiInStart(inHandle);
		}

		if (debug){
			cout << endl << "Debug MidiReceiver():" << endl; 
			cout << "  The unsigned long, result, value was " << result << endl;
			cout << "  MIM_OPEN: " << MIM_OPEN << endl;
			cout << "  MIM_CLOSE: " << MIM_CLOSE << endl;
			cout << "  MIM_DATA:" << MIM_DATA << endl;
			cout << endl; 
		}
/*      
     midiInStop(inHandle);
     midiInReset(inHandle);
     midiInClose(inHandle);
*/ 
	}
}


int main(int argc, char *argv[]) {

	// read config.ini
	// check for user profile

	char * userpath = getenv("USERPROFILE");
	if (userpath != nullptr){
		path = std::string(userpath) + "\\Documents\\Titan\\MidiToWeb\\config.ini";
	} else {
		std::cout << "No user path";
		system ("PAUSE");
        return 1;
	}
	
	// check for config.ini in user profile - if it doesn't exist take the local from program folder
	
	if (!DoesFileExist(path)) path = "config.ini";
	
	INIReader reader(path);
    if (reader.ParseError() < 0) {
		std::cout << "Trying to load " << path << endl;
        std::cout << "Can't load config.ini, terminating program." << endl << endl;
		system ("PAUSE");
        return 1;
    }
	
	// check for correct sections
	if (sections(reader) != "general,mapping,strings," ){
		cout << "Loading " << path << endl;
		std::cout << "Problem with config.ini, terminating program."  << endl << endl;
		system ("PAUSE");
        return 1;
	}
	
	// catch commandline arguments
	for (int i = 1; i < argc; ++i) {
        std::string arg = argv[i];
        if (arg == "-d") {
			debug = true;				// -d for debug mode
        }
		if (arg == "-n") {
			sendRequest = false;		// -n for no requests sent
        }
    }

   int ckey;           // storage for the current keyboard key being pressed
   
   // main program messages
   cout << "--- Titan USB MIDI to WebAPI translator v0.9 (09.06.2018) ---" << endl;
   if (debug) cout << "                === Debug mode ===" << endl;
   if (!sendRequest) cout << "             === No requests sent ===" << endl;
   cout << endl << "Config loaded from " << path << ":" << endl;
   cout << "  ip      = " << reader.Get("general", "ip", "127.0.0.1") << endl;
   cout << "  omni_ch = " << reader.GetInteger("general", "omni_ch", 0)  << endl << endl;
   cout << "Call with option -d for debug mode" << endl;
   cout << "Call with option -n for not sending requests" << endl << "   to only get MIDI messages displayed in debug mode" << endl;
   cout << "Press \"q\" to quit." << endl << endl;

   // register midi events
   MidiReceiver();

   // main program loop
   while (1) {         // event loop
      if (kbhit()) {   // If a key on the computer keyboard has been pressed 
         ckey = getch();
         if (ckey == 'q') break;
      }
   }
   return 0;
}
